@@ -11,6 +11,21 @@
 #include <linux/list.h>
 #include <linux/types.h>
 #include <linux/version.h>
+
+/*
+ * Redhat kernels regularly break the api, so they need their own version
+ * checks. Make sure they always fail for non-rhel kernels.
+ *
+ * By the way, if rhel keeps picking up breaking patches for old majors after
+ * a new one is out, this check could break down and things could get more
+ * complicated. I don't think they do that though...
+ */
+#ifdef RHEL_RELEASE
+#define RHEL_VERSION_GE(a, b)	(RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(a, b))
+#else
+#define RHEL_VERSION_GE(a, b)	0
+#endif
+
 #include "apfs_raw.h"
 
 #define EFSBADCRC	EBADMSG		/* Bad CRC detected */
@@ -29,10 +44,10 @@ static inline bool sb_rdonly(const struct super_block *sb) { return sb->s_flags
 #endif
 
 /* Compatibility wrapper around submit_bh() */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 0, 0)
-#define apfs_submit_bh(op, op_flags, bh) submit_bh(op, op_flags, bh)
-#else
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 0, 0) || RHEL_VERSION_GE(9, 2)
 #define apfs_submit_bh(op, op_flags, bh) submit_bh(op | op_flags, bh)
+#else
+#define apfs_submit_bh(op, op_flags, bh) submit_bh(op, op_flags, bh)
 #endif
 
 /*
@@ -81,7 +96,7 @@ struct apfs_node {
 
 	int key;		/* Offset of the key area in the block */
 	int free;		/* Offset of the free area in the block */
-	int data;		/* Offset of the data area in the block */
+	int val;		/* Offset of the value area in the block */
 
 	int key_free_list_len;	/* Length of the fragmented free key space */
 	int val_free_list_len;	/* Length of the fragmented free value space */
@@ -137,6 +152,7 @@ struct apfs_spaceman {
 	u32 sm_cib_count;		/* Number of chunk-info blocks */
 	u64 sm_free_count;		/* Number of free blocks */
 	u32 sm_addr_offset;		/* Offset of cib addresses in @sm_raw */
+	u64 sm_main_fq_nodes;		/* Number of nodes in the main fq */
 
 	/*
 	 * A range of freed blocks not yet put in the free queue. Extend this as
@@ -165,6 +181,20 @@ struct apfs_spaceman {
 #define APFS_NX_TRANS_COMMITTING	4	/* Commit ongoing */
 #define APFS_NX_TRANS_INCOMPLETE_BLOCK	8	/* A data block is not written in full */
 
+/*
+ * ENOSPC checks before transactions are more or less coarse depending on this.
+ */
+enum apfs_trans_kind {
+	APFS_TRANS_REG,		/* Most transactions */
+	APFS_TRANS_DEL,		/* Transactions that usually free space */
+	APFS_TRANS_SYNC,	/* Sync (or unmount) transaction */
+};
+
+/* The space requirements for each kind of transaction, in blocks */
+#define APFS_REG_ROOM	128
+#define APFS_DEL_ROOM	20
+#define APFS_SYNC_ROOM	6
+
 /*
  * Structure that keeps track of a container transaction.
  */
@@ -194,14 +224,6 @@ struct apfs_bh_info {
 	struct list_head	list;	/* List of buffers in the transaction */
 };
 
-/*
- * Used to report how many operations may be needed for a transaction
- */
-struct apfs_max_ops {
-	int cat;	/* Maximum catalog records that may need changing */
-	int blks;	/* Maximum extent blocks that may need changing */
-};
-
 /*
  * List entry for an in-memory ephemeral object
  */
@@ -228,16 +250,38 @@ struct apfs_ephemeral_object_info {
  */
 #define APFS_FLAGS_SET		4
 
+/*
+ * Wrapper around block devices for portability.
+ */
+struct apfs_blkdev_info {
+	struct block_device *blki_bdev;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 9, 0) || RHEL_VERSION_GE(9, 5)
+	struct file *blki_bdev_file;
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0)
+	struct bdev_handle *blki_bdev_handle;
+#endif
+
+#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 5, 0) && !RHEL_VERSION_GE(9, 4)
+	fmode_t blki_mode;
+#endif
+
+	/*
+	 * For tier 2, we need to remember the mount path for show_options().
+	 * We don't need this for the main device of course, but better to
+	 * keep things consistent.
+	 */
+	char *blki_path;
+};
+
 /*
  * Container superblock data in memory
  */
 struct apfs_nxsb_info {
-	struct block_device *nx_bdev; /* Device for the container */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 9, 0)
-	struct file *nx_bdev_file;
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0)
-	struct bdev_handle *nx_bdev_handle;
-#endif
+	/* Block device info for the container */
+	struct apfs_blkdev_info *nx_blkdev_info;
+	struct apfs_blkdev_info *nx_tier2_info;
+	u64 nx_tier2_bno; /* Offset for tier 2 block numbers */
+
 	struct apfs_nx_superblock *nx_raw; /* On-disk main sb */
 	u64 nx_bno; /* Current block number for the checkpoint superblock */
 	u64 nx_xid; /* Latest transaction id */
@@ -322,6 +366,8 @@ struct apfs_sb_info {
 
 	dev_t s_anon_dev; /* Anonymous device for this volume-snapshot */
 
+	char *s_tier2_path; /* Path to the tier 2 device */
+
 	char *s_snap_name; /* Label for the mounted snapshot */
 	u64 s_snap_xid; /* Transaction id for mounted snapshot */
 
@@ -339,6 +385,7 @@ struct apfs_sb_info {
 
 	struct inode *s_private_dir;	/* Inode for the private directory */
 	struct work_struct s_orphan_cleanup_work;
+	atomic_t s_orphan_cleanup_err;	/* Error from last orphan cleanup */
 };
 
 static inline struct apfs_sb_info *APFS_SB(struct super_block *sb)
@@ -647,8 +694,7 @@ static inline u64 apfs_cat_cnid(struct apfs_key_header *key)
 #define APFS_QUERY_PREV		010000	/* Find previous record */
 
 /*
- * Structure used to retrieve data from an APFS B-Tree. For now only used
- * on the calalog and the object map.
+ * Structure used to retrieve data from an APFS B-Tree.
  */
 struct apfs_query {
 	struct apfs_node *node;		/* Node being searched */
@@ -661,8 +707,8 @@ struct apfs_query {
 	int index;			/* Index of the entry in the node */
 	int key_off;			/* Offset of the key in the node */
 	int key_len;			/* Length of the key */
-	int off;			/* Offset of the data in the node */
-	int len;			/* Length of the data */
+	int off;			/* Offset of the value in the node */
+	int len;			/* Length of the value */
 
 	int depth;			/* Put a limit on recursion */
 };
@@ -926,7 +972,6 @@ extern int apfs_link(struct dentry *old_dentry, struct inode *dir,
 extern int apfs_unlink(struct inode *dir, struct dentry *dentry);
 extern int apfs_rmdir(struct inode *dir, struct dentry *dentry);
 extern int apfs_delete_orphan_link(struct inode *inode);
-extern int APFS_DELETE_ORPHAN_LINK_MAXOPS(void);
 extern u64 apfs_any_orphan_ino(struct super_block *sb, u64 *ino_p);
 
 /* extents.c */
@@ -941,7 +986,6 @@ extern int apfs_flush_extent_cache(struct apfs_dstream_info *dstream);
 extern int apfs_dstream_get_new_bno(struct apfs_dstream_info *dstream, u64 dsblock, u64 *bno);
 extern int apfs_get_new_block(struct inode *inode, sector_t iblock,
 			      struct buffer_head *bh_result, int create);
-extern int APFS_GET_NEW_BLOCK_MAXOPS(void);
 extern int apfs_truncate(struct apfs_dstream_info *dstream, loff_t new_size);
 extern int apfs_inode_delete_front(struct inode *inode);
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0)
@@ -960,7 +1004,7 @@ extern int apfs_fsync(struct file *file, loff_t start, loff_t end, int datasync)
 /* inode.c */
 extern struct inode *apfs_iget(struct super_block *sb, u64 cnid);
 extern int apfs_update_inode(struct inode *inode, char *new_name);
-extern int APFS_UPDATE_INODE_MAXOPS(void);
+extern void apfs_schedule_orphan_cleanup(struct super_block *sb);
 extern void apfs_orphan_cleanup_work(struct work_struct *work);
 extern void apfs_evict_inode(struct inode *inode);
 extern struct inode *apfs_new_inode(struct inode *dir, umode_t mode,
@@ -968,7 +1012,6 @@ extern struct inode *apfs_new_inode(struct inode *dir, umode_t mode,
 extern int apfs_create_inode_rec(struct super_block *sb, struct inode *inode,
 				 struct dentry *dentry);
 extern int apfs_inode_create_exclusive_dstream(struct inode *inode);
-extern int APFS_CREATE_INODE_REC_MAXOPS(void);
 extern int __apfs_write_begin(struct file *file, struct address_space *mapping, loff_t pos, unsigned int len, unsigned int flags, struct page **pagep, void **fsdata);
 extern int __apfs_write_end(struct file *file, struct address_space *mapping, loff_t pos, unsigned int len, unsigned int copied, struct page *page, void *fsdata);
 extern int apfs_dstream_adj_refcnt(struct apfs_dstream_info *dstream, u32 delta);
@@ -1008,7 +1051,6 @@ extern int apfs_getattr(struct mnt_idmap *idmap,
 #endif
 
 extern int apfs_crypto_adj_refcnt(struct super_block *sb, u64 crypto_id, int delta);
-extern int APFS_CRYPTO_ADJ_REFCNT_MAXOPS(void);
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 3, 0)
 extern int apfs_fileattr_get(struct dentry *dentry, struct fileattr *fa);
@@ -1072,6 +1114,7 @@ extern int apfs_free_queue_insert_nocache(struct super_block *sb, u64 bno, u64 c
 extern int apfs_free_queue_insert(struct super_block *sb, u64 bno, u64 count);
 extern int apfs_spaceman_allocate_block(struct super_block *sb, u64 *bno, bool backwards);
 extern int apfs_write_ip_bitmaps(struct super_block *sb);
+extern int apfs_spaceman_get_free_blkcnt(struct super_block *sb, u64 *blkcnt);
 
 /* super.c */
 extern int apfs_map_volume_super_bno(struct super_block *sb, u64 bno, bool check);
@@ -1084,13 +1127,14 @@ extern int apfs_sync_fs(struct super_block *sb, int wait);
 /* transaction.c */
 extern int apfs_cpoint_data_free(struct super_block *sb, u64 bno);
 extern void apfs_transaction_init(struct apfs_nx_transaction *trans);
-extern int apfs_transaction_start(struct super_block *sb, struct apfs_max_ops maxops);
+extern int apfs_transaction_start(struct super_block *sb, enum apfs_trans_kind kind);
 extern int apfs_transaction_commit(struct super_block *sb);
 extern void apfs_inode_join_transaction(struct super_block *sb, struct inode *inode);
 extern int apfs_transaction_join(struct super_block *sb,
 				 struct buffer_head *bh);
 void apfs_transaction_abort(struct super_block *sb);
 extern int apfs_transaction_flush_all_inodes(struct super_block *sb);
+extern int apfs_read_ephemeral_objects(struct super_block *sb);
 
 /* xattr.c */
 extern int ____apfs_xattr_get(struct inode *inode, const char *name, void *buffer,
@@ -1100,7 +1144,6 @@ extern int __apfs_xattr_get(struct inode *inode, const char *name, void *buffer,
 extern int apfs_delete_all_xattrs(struct inode *inode);
 extern int apfs_xattr_set(struct inode *inode, const char *name, const void *value,
 			  size_t size, int flags);
-extern int APFS_XATTR_SET_MAXOPS(void);
 extern ssize_t apfs_listxattr(struct dentry *dentry, char *buffer, size_t size);
 extern int apfs_xattr_get_compressed_data(struct inode *inode, const char *name, struct apfs_compressed_data *cdata);
 extern void apfs_release_compressed_data(struct apfs_compressed_data *cdata);
@@ -1160,8 +1203,29 @@ static inline void apfs_assert_query_is_valid(const struct apfs_query *query)
 static inline void
 apfs_map_bh(struct buffer_head *bh, struct super_block *sb, sector_t block)
 {
+	struct apfs_nxsb_info *nxi = NULL;
+	struct apfs_blkdev_info *info = NULL;
+
+	nxi = APFS_NXI(sb);
+	info = nxi->nx_blkdev_info;
+
+	/*
+	 * Block numbers above s_tier2_bno are for the tier 2 device of a
+	 * fusion drive. Don't bother returning an error code if a regular
+	 * drive gets corrupted and reports a tier 2 block number: just treat
+	 * it the same as any other out-of-range block.
+	 */
+	if (block >= nxi->nx_tier2_bno) {
+		if (nxi->nx_tier2_info) {
+			info = nxi->nx_tier2_info;
+			block -= nxi->nx_tier2_bno;
+		} else {
+			apfs_err(sb, "block number in tier 2 range (0x%llx)", (unsigned long long)block);
+		}
+	}
+
 	set_buffer_mapped(bh);
-	bh->b_bdev = APFS_NXI(sb)->nx_bdev;
+	bh->b_bdev = info->blki_bdev;
 	bh->b_blocknr = block;
 	bh->b_size = sb->s_blocksize;
 }
@@ -1169,20 +1233,59 @@ apfs_map_bh(struct buffer_head *bh, struct super_block *sb, sector_t block)
 static inline struct buffer_head *
 apfs_sb_bread(struct super_block *sb, sector_t block)
 {
-	return __bread_gfp(APFS_NXI(sb)->nx_bdev, block, sb->s_blocksize, __GFP_MOVABLE);
+	struct apfs_nxsb_info *nxi = NULL;
+	struct apfs_blkdev_info *info = NULL;
+
+	nxi = APFS_NXI(sb);
+	info = nxi->nx_blkdev_info;
+
+	if (block >= nxi->nx_tier2_bno) {
+		if (!nxi->nx_tier2_info) {
+			/* Not really a fusion drive, so it's corrupted */
+			apfs_err(sb, "block number in tier 2 range (0x%llx)", (unsigned long long)block);
+			return NULL;
+		}
+		info = nxi->nx_tier2_info;
+		block -= nxi->nx_tier2_bno;
+	}
+
+	return __bread_gfp(info->blki_bdev, block, sb->s_blocksize, __GFP_MOVABLE);
+}
+
+/* Like apfs_getblk(), but doesn't mark the buffer uptodate */
+static inline struct buffer_head *
+__apfs_getblk(struct super_block *sb, sector_t block)
+{
+	struct apfs_nxsb_info *nxi = NULL;
+	struct apfs_blkdev_info *info = NULL;
+
+	nxi = APFS_NXI(sb);
+	info = nxi->nx_blkdev_info;
+
+	if (block >= nxi->nx_tier2_bno) {
+		if (!nxi->nx_tier2_info) {
+			/* Not really a fusion drive, so it's corrupted */
+			apfs_err(sb, "block number in tier 2 range (0x%llx)", (unsigned long long)block);
+			return NULL;
+		}
+		info = nxi->nx_tier2_info;
+		block -= nxi->nx_tier2_bno;
+	}
+
+#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 7, 0)
+	return __getblk_gfp(info->blki_bdev, block, sb->s_blocksize, __GFP_MOVABLE);
+#else
+	return bdev_getblk(info->blki_bdev, block, sb->s_blocksize, __GFP_MOVABLE);
+#endif
 }
 
 /* Use instead of apfs_sb_bread() for blocks that will just be overwritten */
 static inline struct buffer_head *
 apfs_getblk(struct super_block *sb, sector_t block)
 {
-	struct buffer_head *bh;
+	struct buffer_head *bh = NULL;
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 7, 0)
-	bh = __getblk_gfp(APFS_NXI(sb)->nx_bdev, block, sb->s_blocksize, __GFP_MOVABLE);
-#else
-	bh = bdev_getblk(APFS_NXI(sb)->nx_bdev, block, sb->s_blocksize, __GFP_MOVABLE);
-#endif
+	bh = __apfs_getblk(sb, block);
 	if (bh)
 		set_buffer_uptodate(bh);
 	return bh;
