@@ -414,7 +414,6 @@ fail:
 	apfs_free_query(query);
 	return ret;
 }
-#define APFS_CREATE_DENTRY_REC_MAXOPS	1
 
 /**
  * apfs_build_sibling_val - Allocate and initialize a sibling link's value
@@ -491,7 +490,6 @@ fail:
 	apfs_free_query(query);
 	return ret;
 }
-#define APFS_CREATE_SIBLING_LINK_REC_MAXOPS	1
 
 /**
  * apfs_create_sibling_map_rec - Create a sibling map record for a dentry
@@ -534,7 +532,6 @@ fail:
 	apfs_free_query(query);
 	return ret;
 }
-#define APFS_CREATE_SIBLING_MAP_REC_MAXOPS	1
 
 /**
  * apfs_create_sibling_recs - Create sibling link and map records for a dentry
@@ -567,8 +564,6 @@ static int apfs_create_sibling_recs(struct dentry *dentry,
 	*sibling_id = cnid;
 	return 0;
 }
-#define APFS_CREATE_SIBLING_RECS_MAXOPS	(APFS_CREATE_SIBLING_LINK_REC_MAXOPS + \
-					 APFS_CREATE_SIBLING_MAP_REC_MAXOPS)
 
 /**
  * apfs_create_dentry - Create all records for a new dentry
@@ -613,9 +608,6 @@ static int apfs_create_dentry(struct dentry *dentry, struct inode *inode)
 	apfs_inode_join_transaction(parent->i_sb, parent);
 	return 0;
 }
-#define APFS_CREATE_DENTRY_MAXOPS	(APFS_CREATE_SIBLING_RECS_MAXOPS + \
-					 APFS_CREATE_DENTRY_REC_MAXOPS + \
-					 APFS_UPDATE_INODE_MAXOPS())
 
 /**
  * apfs_undo_create_dentry - Clean up apfs_create_dentry()
@@ -643,15 +635,9 @@ int apfs_mkany(struct inode *dir, struct dentry *dentry, umode_t mode,
 {
 	struct super_block *sb = dir->i_sb;
 	struct inode *inode;
-	struct apfs_max_ops maxops;
 	int err;
 
-	maxops.cat = APFS_CREATE_INODE_REC_MAXOPS() + APFS_CREATE_DENTRY_MAXOPS;
-	if (symname)
-		maxops.cat += APFS_XATTR_SET_MAXOPS();
-	maxops.blks = 0;
-
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 
@@ -818,8 +804,6 @@ static int apfs_prepare_dentry_for_link(struct dentry *dentry)
 	return apfs_create_dentry_rec(d_inode(dentry), &dentry->d_name,
 				      apfs_ino(parent), sibling_id);
 }
-#define APFS_PREPARE_DENTRY_FOR_LINK_MAXOPS	(1 + APFS_CREATE_SIBLING_RECS_MAXOPS + \
-						 APFS_CREATE_DENTRY_REC_MAXOPS)
 
 /**
  * __apfs_undo_link - Clean up __apfs_link()
@@ -875,22 +859,15 @@ fail:
 	drop_nlink(inode);
 	return err;
 }
-#define __APFS_LINK_MAXOPS	(APFS_UPDATE_INODE_MAXOPS() + \
-				 APFS_PREPARE_DENTRY_FOR_LINK_MAXOPS + \
-				 APFS_CREATE_DENTRY_MAXOPS)
 
 int apfs_link(struct dentry *old_dentry, struct inode *dir,
 	      struct dentry *dentry)
 {
 	struct super_block *sb = dir->i_sb;
 	struct inode *inode = d_inode(old_dentry);
-	struct apfs_max_ops maxops;
 	int err;
 
-	maxops.cat = __APFS_LINK_MAXOPS;
-	maxops.blks = 0;
-
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 
@@ -954,7 +931,6 @@ fail:
 	apfs_free_query(query);
 	return ret;
 }
-#define APFS_DELETE_SIBLING_LINK_REC_MAXOPS	1
 
 /**
  * apfs_delete_sibling_map_rec - Delete the sibling map record for a dentry
@@ -995,7 +971,6 @@ fail:
 	apfs_free_query(query);
 	return ret;
 }
-#define APFS_DELETE_SIBLING_MAP_REC_MAXOPS	1
 
 /**
  * apfs_delete_dentry - Delete all records for a dentry
@@ -1048,9 +1023,6 @@ static int apfs_delete_dentry(struct dentry *dentry)
 	apfs_inode_join_transaction(sb, parent);
 	return err;
 }
-#define APFS_DELETE_DENTRY_MAXOPS	(1 + APFS_DELETE_SIBLING_LINK_REC_MAXOPS + \
-					 APFS_DELETE_SIBLING_MAP_REC_MAXOPS + \
-					 APFS_UPDATE_INODE_MAXOPS())
 
 /**
  * apfs_undo_delete_dentry - Clean up apfs_delete_dentry()
@@ -1217,8 +1189,6 @@ fail:
 	kfree(qname.name);
 	return err;
 }
-#define APFS_CREATE_ORPHAN_LINK_MAXOPS	(APFS_CREATE_DENTRY_REC_MAXOPS + \
-					 APFS_UPDATE_INODE_MAXOPS())
 
 /**
  * apfs_delete_orphan_link - Delete the link for an orphan inode
@@ -1269,10 +1239,6 @@ fail:
 	kfree(qname.name);
 	return err;
 }
-int APFS_DELETE_ORPHAN_LINK_MAXOPS(void)
-{
-	return 1 + APFS_UPDATE_INODE_MAXOPS();
-}
 
 /**
  * __apfs_undo_unlink - Clean up __apfs_unlink()
@@ -1368,20 +1334,13 @@ fail:
 		__apfs_undo_unlink(dentry);
 	return err;
 }
-#define __APFS_UNLINK_MAXOPS	(APFS_DELETE_DENTRY_MAXOPS + \
-				 APFS_CREATE_ORPHAN_LINK_MAXOPS + \
-				 APFS_UPDATE_INODE_MAXOPS())
 
 int apfs_unlink(struct inode *dir, struct dentry *dentry)
 {
 	struct super_block *sb = dir->i_sb;
-	struct apfs_max_ops maxops;
 	int err;
 
-	maxops.cat = __APFS_UNLINK_MAXOPS;
-	maxops.blks = 0;
-
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_DEL);
 	if (err)
 		return err;
 
@@ -1427,7 +1386,6 @@ int apfs_rename(struct mnt_idmap *idmap, struct inode *old_dir,
 	struct super_block *sb = old_dir->i_sb;
 	struct inode *old_inode = d_inode(old_dentry);
 	struct inode *new_inode = d_inode(new_dentry);
-	struct apfs_max_ops maxops;
 	int err;
 
 	if (new_inode && APFS_I(new_inode)->i_nchildren)
@@ -1436,12 +1394,7 @@ int apfs_rename(struct mnt_idmap *idmap, struct inode *old_dir,
 	if (flags & ~RENAME_NOREPLACE) /* TODO: support RENAME_EXCHANGE */
 		return -EINVAL;
 
-	maxops.cat = __APFS_UNLINK_MAXOPS + __APFS_LINK_MAXOPS;
-	if (new_inode)
-		maxops.cat += __APFS_UNLINK_MAXOPS;
-	maxops.blks = 0;
-
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 
