@@ -10,7 +10,7 @@
 #include <linux/blk_types.h>
 #include "apfs.h"
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0) || RHEL_VERSION_GE(9, 3)
 #include <linux/sched/mm.h>
 #endif
 
@@ -20,7 +20,7 @@
 
 #define MAX_PFK_LEN	512
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0) || RHEL_VERSION_GE(9, 3)
 
 static int apfs_read_folio(struct file *file, struct folio *folio)
 {
@@ -90,7 +90,6 @@ out:
 	apfs_free_query(query);
 	return ret;
 }
-#define APFS_CREATE_DSTREAM_REC_MAXOPS	1
 
 static int apfs_check_dstream_refcnt(struct inode *inode);
 static int apfs_put_dstream_rec(struct apfs_dstream_info *dstream);
@@ -308,7 +307,6 @@ out:
 	apfs_free_query(query);
 	return ret;
 }
-#define APFS_CREATE_CRYPTO_REC_MAXOPS	1
 
 /**
  * apfs_dflt_key_class - Returns default key class for files in volume
@@ -371,10 +369,6 @@ out:
 	apfs_free_query(query);
 	return ret;
 }
-int APFS_CRYPTO_ADJ_REFCNT_MAXOPS(void)
-{
-	return 1;
-}
 
 /**
  * apfs_crypto_set_key - Modify content of crypto state record
@@ -423,7 +417,6 @@ out:
 	apfs_free_query(query);
 	return ret;
 }
-#define APFS_CRYPTO_SET_KEY_MAXOPS	1
 
 /**
  * apfs_crypto_get_key - Retrieve content of crypto state record
@@ -502,7 +495,7 @@ int __apfs_write_begin(struct file *file, struct address_space *mapping, loff_t
 		}
 	}
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0) || RHEL_VERSION_GE(9, 3)
 	flags = memalloc_nofs_save();
 	page = grab_cache_page_write_begin(mapping, index);
 	memalloc_nofs_restore(flags);
@@ -588,7 +581,7 @@ out_put_page:
 static int apfs_write_begin(struct file *file, struct address_space *mapping,
 			    loff_t pos, unsigned int len,
 			    struct folio **foliop, void **fsdata)
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0) || RHEL_VERSION_GE(9, 3)
 static int apfs_write_begin(struct file *file, struct address_space *mapping,
 			    loff_t pos, unsigned int len,
 			    struct page **pagep, void **fsdata)
@@ -604,23 +597,15 @@ static int apfs_write_begin(struct file *file, struct address_space *mapping,
 	struct page *page = NULL;
 	struct page **pagep = &page;
 #endif
-	int blkcount = (len + sb->s_blocksize - 1) >> inode->i_blkbits;
-	struct apfs_max_ops maxops;
 	int err;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0) || RHEL_VERSION_GE(9, 3)
 	unsigned int flags = 0;
 #endif
 
 	if (unlikely(pos >= APFS_MAX_FILE_SIZE))
 		return -EFBIG;
 
-	maxops.cat = APFS_CREATE_DSTREAM_REC_MAXOPS +
-		     APFS_CREATE_CRYPTO_REC_MAXOPS +
-		     APFS_UPDATE_INODE_MAXOPS() +
-		     blkcount * APFS_GET_NEW_BLOCK_MAXOPS();
-	maxops.blks = blkcount;
-
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 
@@ -698,23 +683,23 @@ fail:
 	return err;
 }
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 18, 0)
-static void apfs_noop_invalidatepage(struct page *page, unsigned int offset, unsigned int length)
-#else
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0) || RHEL_VERSION_GE(9, 3)
 static void apfs_noop_invalidate_folio(struct folio *folio, size_t offset, size_t length)
+#else
+static void apfs_noop_invalidatepage(struct page *page, unsigned int offset, unsigned int length)
 #endif
 {
 }
 
 /* bmap is not implemented to avoid issues with CoW on swapfiles */
 static const struct address_space_operations apfs_aops = {
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0) || RHEL_VERSION_GE(9, 2)
 	.dirty_folio	= block_dirty_folio,
 #elif LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)
 	.set_page_dirty	= __set_page_dirty_buffers,
 #endif
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 19, 0) || RHEL_VERSION_GE(9, 3)
 	.read_folio	= apfs_read_folio,
 #else
 	.readpage	= apfs_readpage,
@@ -730,10 +715,10 @@ static const struct address_space_operations apfs_aops = {
 	.write_end	= apfs_write_end,
 
 	/* The intention is to keep bhs around until the transaction is over */
-#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 18, 0)
-	.invalidatepage	= apfs_noop_invalidatepage,
-#else
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0) || RHEL_VERSION_GE(9, 3)
 	.invalidate_folio = apfs_noop_invalidate_folio,
+#else
+	.invalidatepage	= apfs_noop_invalidatepage,
 #endif
 };
 
@@ -1253,7 +1238,6 @@ fail:
 	kfree(new_val);
 	return err;
 }
-#define APFS_INODE_RENAME_MAXOPS	1
 
 /**
  * apfs_create_dstream_xfield - Create the inode xfield for a new data stream
@@ -1308,7 +1292,6 @@ fail:
 	kfree(new_val);
 	return err;
 }
-#define APFS_CREATE_DSTREAM_XFIELD_MAXOPS	1
 
 /**
  * apfs_inode_resize - Update the sizes reported in an inode record
@@ -1359,7 +1342,6 @@ static int apfs_inode_resize(struct inode *inode, struct apfs_query *query)
 	/* This inode has no dstream xfield, so we need to create it */
 	return apfs_create_dstream_xfield(inode, query);
 }
-#define APFS_INODE_RESIZE_MAXOPS	(1 + APFS_CREATE_DSTREAM_XFIELD_MAXOPS)
 
 /**
  * apfs_create_sparse_xfield - Create an inode xfield to count sparse bytes
@@ -1563,10 +1545,6 @@ fail:
 	apfs_free_query(query);
 	return err;
 }
-int APFS_UPDATE_INODE_MAXOPS(void)
-{
-	return APFS_INODE_RENAME_MAXOPS + APFS_INODE_RESIZE_MAXOPS + 1;
-}
 
 /**
  * apfs_delete_inode - Delete an inode record
@@ -1650,7 +1628,6 @@ static int apfs_delete_inode(struct inode *inode)
 	ai->i_cleaned = true;
 	return ret;
 }
-#define APFS_DELETE_INODE_MAXOPS	1
 
 /**
  * apfs_clean_single_orphan - Clean the given orphan file
@@ -1662,12 +1639,11 @@ static int apfs_delete_inode(struct inode *inode)
 static int apfs_clean_single_orphan(struct inode *inode)
 {
 	struct super_block *sb = inode->i_sb;
-	struct apfs_max_ops maxops = {0}; /* TODO: rethink this stuff */
 	u64 ino = apfs_ino(inode);
 	bool eagain = false;
 	int err;
 
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_DEL);
 	if (err)
 		return err;
 	err = apfs_delete_inode(inode);
@@ -1746,6 +1722,32 @@ out:
 	return err;
 }
 
+/**
+ * apfs_schedule_orphan_cleanup - Schedule cleanup for orphan inodes
+ * @sb: filesystem superblock
+ */
+void apfs_schedule_orphan_cleanup(struct super_block *sb)
+{
+	struct apfs_sb_info *sbi = APFS_SB(sb);
+
+	/*
+	 * Don't schedule cleanups during unmount: completing all of it could
+	 * take a while so just leave future mounts to handle the orphans.
+	 */
+	if (atomic_read(&sb->s_active) == 0)
+		return;
+
+	/*
+	 * Don't keep retrying orphan cleanups nonstop when they run into an
+	 * unexpected error: it won't do any good and it will flood dmesg. We
+	 * will retry eventually for ENOSPC, but that's handled elsewhere.
+	 */
+	if (atomic_read(&sbi->s_orphan_cleanup_err))
+		return;
+
+	schedule_work(&sbi->s_orphan_cleanup_work);
+}
+
 /**
  * apfs_clean_orphans - Delete as many orphan files as is reasonable
  * @sb: filesystem superblock
@@ -1754,26 +1756,27 @@ out:
  */
 static int apfs_clean_orphans(struct super_block *sb)
 {
+	struct apfs_sb_info *sbi = APFS_SB(sb);
 	int ret, i;
 
 	for (i = 0; i < 100; ++i) {
 		ret = apfs_clean_any_orphan(sb);
-		if (ret) {
-			if (ret == -ENODATA)
-				return 0;
-			if (ret == -EAGAIN)
-				break;
-			apfs_err(sb, "failed to delete an orphan file");
-			return ret;
-		}
+		if (ret == 0)
+			continue;
+		if (ret == -ENODATA)
+			return 0;
+		if (ret == -EAGAIN)
+			break;
+		apfs_err(sb, "failed to delete an orphan file");
+		atomic_set(&sbi->s_orphan_cleanup_err, ret);
+		return ret;
 	}
 
 	/*
 	 * If a file is too big, or if there are too many files, take a break
 	 * and continue later.
 	 */
-	if (atomic_read(&sb->s_active) != 0)
-		schedule_work(&APFS_SB(sb)->s_orphan_cleanup_work);
+	apfs_schedule_orphan_cleanup(sb);
 	return 0;
 }
 
@@ -1789,18 +1792,15 @@ void apfs_evict_inode(struct inode *inode)
 	if (!ai->i_has_dstream || ai->i_dstream.ds_size == 0) {
 		/* For files with no extents, scheduled cleanup wastes time */
 		err = apfs_clean_single_orphan(inode);
-		if (err)
+		if (err) {
 			apfs_err(sb, "failed to clean orphan 0x%llx (err:%d)", apfs_ino(inode), err);
+			atomic_set(&APFS_SB(sb)->s_orphan_cleanup_err, err);
+		}
 		goto out;
 	}
 
-	/*
-	 * If the inode still has extents then schedule cleanup for the rest
-	 * of it. Not during unmount though: completing all cleanup could take
-	 * a while so just leave future mounts to handle the orphans.
-	 */
-	if (atomic_read(&sb->s_active))
-		schedule_work(&APFS_SB(sb)->s_orphan_cleanup_work);
+	/* If the inode still has extents then schedule cleanup for the rest */
+	apfs_schedule_orphan_cleanup(sb);
 out:
 	truncate_inode_pages_final(&inode->i_data);
 	clear_inode(inode);
@@ -1979,10 +1979,6 @@ fail:
 	apfs_free_query(query);
 	return ret;
 }
-int APFS_CREATE_INODE_REC_MAXOPS(void)
-{
-	return 1;
-}
 
 /**
  * apfs_setsize - Change the size of a regular file
@@ -2037,12 +2033,14 @@ int apfs_setattr(struct mnt_idmap *idmap,
 {
 	struct inode *inode = d_inode(dentry);
 	struct super_block *sb = inode->i_sb;
-	struct apfs_max_ops maxops;
 	bool resizing = S_ISREG(inode->i_mode) && (iattr->ia_valid & ATTR_SIZE);
+	bool shrinking = false;
 	int err;
 
 	if (resizing && iattr->ia_size > APFS_MAX_FILE_SIZE)
 		return -EFBIG;
+	if (resizing && iattr->ia_size < inode->i_size)
+		shrinking = true;
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(5, 12, 0)
 	err = setattr_prepare(dentry, iattr);
@@ -2054,11 +2052,8 @@ int apfs_setattr(struct mnt_idmap *idmap,
 	if (err)
 		return err;
 
-	maxops.cat = APFS_UPDATE_INODE_MAXOPS();
-	maxops.blks = 0;
-
 	/* TODO: figure out why ->write_inode() isn't firing */
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, shrinking ? APFS_TRANS_DEL : APFS_TRANS_REG);
 	if (err)
 		return err;
 	apfs_inode_join_transaction(sb, inode);
@@ -2098,13 +2093,9 @@ int apfs_update_time(struct inode *inode, int flags)
 #endif
 {
 	struct super_block *sb = inode->i_sb;
-	struct apfs_max_ops maxops;
 	int err;
 
-	maxops.cat = APFS_UPDATE_INODE_MAXOPS();
-	maxops.blks = 0;
-
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 	apfs_inode_join_transaction(sb, inode);
@@ -2165,7 +2156,6 @@ static int apfs_ioc_set_dir_class(struct file *file, u32 __user *user_class)
 	struct inode *inode = file_inode(file);
 	struct apfs_inode_info *ai = APFS_I(inode);
 	struct super_block *sb = inode->i_sb;
-	struct apfs_max_ops maxops;
 	u32 class;
 	int err;
 
@@ -2174,10 +2164,7 @@ static int apfs_ioc_set_dir_class(struct file *file, u32 __user *user_class)
 
 	ai->i_key_class = class;
 
-	maxops.cat = APFS_UPDATE_INODE_MAXOPS();
-	maxops.blks = 0;
-
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 	apfs_inode_join_transaction(sb, inode);
@@ -2199,7 +2186,6 @@ static int apfs_ioc_set_pfk(struct file *file, void __user *user_pfk)
 	struct apfs_crypto_state_val *pfk;
 	struct apfs_inode_info *ai = APFS_I(inode);
 	struct apfs_dstream_info *dstream = &ai->i_dstream;
-	struct apfs_max_ops maxops;
 	unsigned int key_len, key_class;
 	int err;
 
@@ -2217,10 +2203,7 @@ static int apfs_ioc_set_pfk(struct file *file, void __user *user_pfk)
 	}
 	pfk->refcnt = cpu_to_le32(1);
 
-	maxops.cat = APFS_CRYPTO_SET_KEY_MAXOPS + APFS_UPDATE_INODE_MAXOPS();
-	maxops.blks = 0;
-
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err) {
 		kfree(pfk);
 		return err;
@@ -2387,7 +2370,6 @@ static int apfs_ioc_getflags(struct file *file, int __user *arg)
 static int apfs_do_ioc_setflags(struct inode *inode, unsigned int newflags)
 {
 	struct super_block *sb = inode->i_sb;
-	struct apfs_max_ops maxops;
 	unsigned int oldflags;
 	int err;
 
@@ -2398,9 +2380,7 @@ static int apfs_do_ioc_setflags(struct inode *inode, unsigned int newflags)
 	if (err)
 		return err;
 
-	maxops.cat = APFS_UPDATE_INODE_MAXOPS();
-	maxops.blks = 0;
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 
@@ -2470,7 +2450,6 @@ int apfs_fileattr_set(struct user_namespace *mnt_userns, struct dentry *dentry,
 {
 	struct inode *inode = d_inode(dentry);
 	struct super_block *sb = inode->i_sb;
-	struct apfs_max_ops maxops;
 	int err;
 
 	if (sb->s_flags & SB_RDONLY)
@@ -2483,9 +2462,7 @@ int apfs_fileattr_set(struct user_namespace *mnt_userns, struct dentry *dentry,
 
 	lockdep_assert_held_write(&inode->i_rwsem);
 
-	maxops.cat = APFS_UPDATE_INODE_MAXOPS();
-	maxops.blks = 0;
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 
@@ -2513,7 +2490,6 @@ int apfs_fileattr_set(struct mnt_idmap *idmap, struct dentry *dentry, struct fil
 {
 	struct inode *inode = d_inode(dentry);
 	struct super_block *sb = inode->i_sb;
-	struct apfs_max_ops maxops;
 	int err;
 
 	if (sb->s_flags & SB_RDONLY)
@@ -2526,9 +2502,7 @@ int apfs_fileattr_set(struct mnt_idmap *idmap, struct dentry *dentry, struct fil
 
 	lockdep_assert_held_write(&inode->i_rwsem);
 
-	maxops.cat = APFS_UPDATE_INODE_MAXOPS();
-	maxops.blks = 0;
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 
