@@ -395,13 +395,11 @@ static int apfs_do_ioc_take_snapshot(struct inode *mntpoint, const char *name, i
 	struct apfs_sb_info *sbi = APFS_SB(sb);
 	struct apfs_superblock *vsb_raw = NULL;
 	struct apfs_omap *omap = sbi->s_omap;
-	/* TODO: remember to update the maxops in the future */
-	struct apfs_max_ops maxops = {0};
 	u64 sblock_oid;
 	bool eexist = false;
 	int err;
 
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 
@@ -420,6 +418,7 @@ static int apfs_do_ioc_take_snapshot(struct inode *mntpoint, const char *name, i
 		err = apfs_transaction_commit(sb);
 		if (err)
 			goto fail;
+		apfs_warn(sb, "snapshot name already in use (%s)", name);
 		return -EEXIST;
 	}
 
@@ -585,15 +584,22 @@ fail:
 
 static int apfs_snap_sblock_from_query(struct apfs_query *query, u64 *sblock_oid)
 {
+	struct super_block *sb = query->node->object.sb;
 	char *raw = query->node->object.data;
 	struct apfs_snap_metadata_val *val = NULL;
 
 	if (query->len < sizeof(*val)) {
-		apfs_err(query->node->object.sb, "bad value length (%d)", query->len);
+		apfs_err(sb, "bad value length (%d)", query->len);
 		return -EFSCORRUPTED;
 	}
 	val = (struct apfs_snap_metadata_val *)(raw + query->off);
 
+	/* I don't know anything about these snapshots, can they be mounted? */
+	if (le32_to_cpu(val->flags) & APFS_SNAP_META_PENDING_DATALESS) {
+		apfs_warn(sb, "the requested snapshot is dataless");
+		return -EOPNOTSUPP;
+	}
+
 	*sblock_oid = le64_to_cpu(val->sblock_oid);
 	return 0;
 }
