@@ -57,7 +57,7 @@ static struct apfs_nxsb_info *apfs_nx_find_by_dev(dev_t dev)
 
 	lockdep_assert_held(&nxs_mutex);
 	list_for_each_entry(curr, &nxs, nx_list) {
-		struct block_device *curr_bdev = curr->nx_bdev;
+		struct block_device *curr_bdev = curr->nx_blkdev_info->blki_bdev;
 
 		if (curr_bdev->bd_dev == dev)
 			return curr;
@@ -65,6 +65,22 @@ static struct apfs_nxsb_info *apfs_nx_find_by_dev(dev_t dev)
 	return NULL;
 }
 
+/**
+ * apfs_blkdev_set_blocksize - Set the blocksize for a block device
+ * @info:	info struct for the block device
+ * @size:	size to set
+ *
+ * Returns 0 on success, or a negative error code in case of failure.
+ */
+static int apfs_blkdev_set_blocksize(struct apfs_blkdev_info *info, int size)
+{
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 10, 0)
+	return set_blocksize(info->blki_bdev_file, size);
+#else
+	return set_blocksize(info->blki_bdev, size);
+#endif
+}
+
 /**
  * apfs_sb_set_blocksize - Set the block size for the container's device
  * @sb:		superblock structure
@@ -75,11 +91,7 @@ static struct apfs_nxsb_info *apfs_nx_find_by_dev(dev_t dev)
  */
 static int apfs_sb_set_blocksize(struct super_block *sb, int size)
 {
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 10, 0)
-	if (set_blocksize(APFS_NXI(sb)->nx_bdev_file, size))
-#else
-	if (set_blocksize(APFS_NXI(sb)->nx_bdev, size))
-#endif
+	if (apfs_blkdev_set_blocksize(APFS_NXI(sb)->nx_blkdev_info, size))
 		return 0;
 	sb->s_blocksize = size;
 	sb->s_blocksize_bits = blksize_bits(size);
@@ -115,6 +127,12 @@ static struct buffer_head *apfs_read_super_copy(struct super_block *sb)
 	msb_raw = (struct apfs_nx_superblock *)bh->b_data;
 	blocksize = le32_to_cpu(msb_raw->nx_block_size);
 
+	sb->s_magic = le32_to_cpu(msb_raw->nx_magic);
+	if (sb->s_magic != APFS_NX_MAGIC) {
+		apfs_warn(sb, "not an apfs container - are you mounting the right partition?");
+		goto fail;
+	}
+
 	if (sb->s_blocksize != blocksize) {
 		brelse(bh);
 
@@ -130,11 +148,6 @@ static struct buffer_head *apfs_read_super_copy(struct super_block *sb)
 		msb_raw = (struct apfs_nx_superblock *)bh->b_data;
 	}
 
-	sb->s_magic = le32_to_cpu(msb_raw->nx_magic);
-	if (sb->s_magic != APFS_NX_MAGIC) {
-		apfs_err(sb, "not an apfs filesystem");
-		goto fail;
-	}
 	if (!apfs_obj_verify_csum(sb, bh))
 		apfs_notice(sb, "backup superblock seems corrupted");
 	return bh;
@@ -181,6 +194,80 @@ out_unlock:
 static int apfs_check_nx_features(struct super_block *sb);
 static void apfs_set_trans_buffer_limit(struct super_block *sb);
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0)
+static inline void import_uuid(uuid_t *dst, const __u8 *src)
+{
+	memcpy(dst, src, sizeof(uuid_t));
+}
+#endif
+
+/**
+ * apfs_check_fusion_uuid - Verify that the main and tier 2 devices match
+ * @sb: filesystem superblock
+ *
+ * Returns 0 on success, or a negative error code in case of failure.
+ */
+static int apfs_check_fusion_uuid(struct super_block *sb)
+{
+	struct apfs_nxsb_info *nxi = NULL;
+	struct apfs_nx_superblock *main_raw = NULL, *tier2_raw = NULL;
+	uuid_t main_uuid, tier2_uuid;
+	struct buffer_head *bh = NULL;
+
+	nxi = APFS_NXI(sb);
+	main_raw = nxi->nx_raw;
+	if (!main_raw) {
+		apfs_alert(sb, "fusion uuid checks are misplaced");
+		return -EINVAL;
+	}
+	import_uuid(&main_uuid, main_raw->nx_fusion_uuid);
+	main_raw = NULL;
+
+	if (!nxi->nx_tier2_info) {
+		/* Not a fusion drive */
+		if (!uuid_is_null(&main_uuid)) {
+			apfs_err(sb, "fusion uuid on a regular drive");
+			return -EFSCORRUPTED;
+		}
+		return 0;
+	}
+	if (uuid_is_null(&main_uuid)) {
+		apfs_err(sb, "no fusion uuid on fusion drive");
+		return -EFSCORRUPTED;
+	}
+
+	/* Tier 2 also has a copy of the superblock in block zero */
+	bh = apfs_sb_bread(sb, nxi->nx_tier2_bno);
+	if (IS_ERR(bh))
+		return PTR_ERR(bh);
+	tier2_raw = (struct apfs_nx_superblock *)bh->b_data;
+	import_uuid(&tier2_uuid, tier2_raw->nx_fusion_uuid);
+	brelse(bh);
+	bh = NULL;
+	tier2_raw = NULL;
+
+	/*
+	 * The only difference between both superblocks (other than the
+	 * checksum) is this one bit here, so it can be used to tell which is
+	 * main and which is tier 2. By the way, the reference seems to have
+	 * this backwards.
+	 */
+	if (main_uuid.b[15] & 0x01) {
+		apfs_warn(sb, "bad bit on main device - are you mixing up main and tier 2?");
+		return -EINVAL;
+	}
+	if (!(tier2_uuid.b[15] & 0x01)) {
+		apfs_warn(sb, "bad bit on tier 2 device - are you mixing up main and tier 2?");
+		return -EINVAL;
+	}
+	tier2_uuid.b[15] &= ~0x01;
+	if (!uuid_equal(&main_uuid, &tier2_uuid)) {
+		apfs_warn(sb, "the devices are not part of the same fusion drive");
+		return -EINVAL;
+	}
+	return 0;
+}
+
 /**
  * apfs_read_main_super - Find the container superblock and read it into memory
  * @sb:	superblock structure
@@ -211,6 +298,13 @@ static int apfs_read_main_super(struct super_block *sb)
 		goto out;
 	}
 
+	/*
+	 * We won't know the block size until we read the backup superblock,
+	 * so we can't set this up correctly yet. But we do know that the
+	 * backup superblock itself is always in the main device.
+	 */
+	nxi->nx_tier2_bno = APFS_NX_BLOCK_NUM + 1;
+
 	/* Read the superblock from the last clean unmount */
 	bh = apfs_read_super_copy(sb);
 	if (IS_ERR(bh)) {
@@ -220,6 +314,13 @@ static int apfs_read_main_super(struct super_block *sb)
 	}
 	msb_raw = (struct apfs_nx_superblock *)bh->b_data;
 
+	/*
+	 * Now that we confirmed the block size, we can set this up for real.
+	 * It's important to do this early because I don't know which mount
+	 * objects could get moved to tier 2.
+	 */
+	nxi->nx_tier2_bno = APFS_FUSION_TIER2_DEVICE_BYTE_ADDR >> sb->s_blocksize_bits;
+
 	/* We want to mount the latest valid checkpoint among the descriptors */
 	desc_base = le64_to_cpu(msb_raw->nx_xp_desc_base);
 	if (desc_base >> 63 != 0) {
@@ -277,6 +378,19 @@ static int apfs_read_main_super(struct super_block *sb)
 	apfs_set_trans_buffer_limit(sb);
 
 	err = apfs_check_nx_features(sb);
+	if (err)
+		goto out;
+
+	/*
+	 * This check is technically too late: if main and tier 2 are backwards
+	 * then we have already attempted (and failed) to read the checkpoint
+	 * from tier 2. This may lead to a confusing error message if tier 2
+	 * is absurdly tiny, not a big deal.
+	 */
+	err = apfs_check_fusion_uuid(sb);
+	if (err)
+		goto out;
+
 out:
 	brelse(bh);
 	mutex_unlock(&nxs_mutex);
@@ -310,6 +424,33 @@ static void apfs_update_software_info(struct super_block *sb)
 
 static struct file_system_type apfs_fs_type;
 
+/**
+ * apfs_blkdev_cleanup - Clean up after a block device
+ * @info:	info struct to clean up
+ */
+static void apfs_blkdev_cleanup(struct apfs_blkdev_info *info)
+{
+	if (!info)
+		return;
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 9, 0) || RHEL_VERSION_GE(9, 5)
+	fput(info->blki_bdev_file);
+	info->blki_bdev_file = NULL;
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0)
+	bdev_release(info->blki_bdev_handle);
+	info->blki_bdev_handle = NULL;
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0) || RHEL_VERSION_GE(9, 4)
+	blkdev_put(info->blki_bdev, &apfs_fs_type);
+#else
+	blkdev_put(info->blki_bdev, info->blki_mode);
+#endif
+	info->blki_bdev = NULL;
+
+	kfree(info->blki_path);
+	info->blki_path = NULL;
+	kfree(info);
+}
+
 /**
  * apfs_free_main_super - Clean up apfs_read_main_super()
  * @sbi:	in-memory superblock info
@@ -319,19 +460,11 @@ static struct file_system_type apfs_fs_type;
 static inline void apfs_free_main_super(struct apfs_sb_info *sbi)
 {
 	struct apfs_nxsb_info *nxi = sbi->s_nxi;
-#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 5, 0)
-	fmode_t mode = FMODE_READ | FMODE_EXCL;
-#endif
 	struct apfs_ephemeral_object_info *eph_list = NULL;
 	struct apfs_spaceman *sm = NULL;
 	u32 bmap_idx;
 	int i;
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 5, 0)
-	if (nxi->nx_flags & APFS_READWRITE)
-		mode |= FMODE_WRITE;
-#endif
-
 	mutex_lock(&nxs_mutex);
 
 	list_del(&sbi->list);
@@ -353,15 +486,10 @@ static inline void apfs_free_main_super(struct apfs_sb_info *sbi)
 	kfree(nxi->nx_raw);
 	nxi->nx_raw = NULL;
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 9, 0)
-	fput(nxi->nx_bdev_file);
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0)
-	bdev_release(nxi->nx_bdev_handle);
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0)
-	blkdev_put(nxi->nx_bdev, &apfs_fs_type);
-#else
-	blkdev_put(nxi->nx_bdev, mode);
-#endif
+	apfs_blkdev_cleanup(nxi->nx_blkdev_info);
+	nxi->nx_blkdev_info = NULL;
+	apfs_blkdev_cleanup(nxi->nx_tier2_info);
+	nxi->nx_tier2_info = NULL;
 
 	list_del(&nxi->nx_list);
 	sm = nxi->nx_spaceman;
@@ -784,10 +912,9 @@ static void apfs_put_super(struct super_block *sb)
 	if (!(sb->s_flags & SB_RDONLY)) {
 		struct apfs_superblock *vsb_raw;
 		struct buffer_head *vsb_bh;
-		struct apfs_max_ops maxops = {0};
 		int err;
 
-		err = apfs_transaction_start(sb, maxops);
+		err = apfs_transaction_start(sb, APFS_TRANS_SYNC);
 		if (err) {
 			apfs_err(sb, "unmount transaction start failed (err:%d)", err);
 			goto fail;
@@ -840,7 +967,7 @@ static struct inode *apfs_alloc_inode(struct super_block *sb)
 	struct apfs_inode_info *ai;
 	struct apfs_dstream_info *dstream;
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0) || RHEL_VERSION_GE(9, 1)
 	ai = alloc_inode_sb(sb, apfs_inode_cachep, GFP_KERNEL);
 #else
 	ai = kmem_cache_alloc(apfs_inode_cachep, GFP_KERNEL);
@@ -904,13 +1031,9 @@ static int apfs_write_inode(struct inode *inode, struct writeback_control *wbc)
 {
 	struct super_block *sb = inode->i_sb;
 	struct apfs_nxsb_info *nxi = APFS_SB(sb)->s_nxi;
-	struct apfs_max_ops maxops;
 	int err;
 
-	maxops.cat = APFS_UPDATE_INODE_MAXOPS();
-	maxops.blks = 0;
-
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_REG);
 	if (err)
 		return err;
 	err = apfs_update_inode(inode, NULL /* new_name */);
@@ -938,85 +1061,6 @@ static void destroy_inodecache(void)
 	kmem_cache_destroy(apfs_inode_cachep);
 }
 
-/**
- * apfs_count_used_blocks - Count the blocks in use across all volumes
- * @sb:		filesystem superblock
- * @count:	on return it will store the block count
- *
- * This function probably belongs in a separate file, but for now it is
- * only called by statfs.
- */
-static int apfs_count_used_blocks(struct super_block *sb, u64 *count)
-{
-	struct apfs_nx_superblock *msb_raw = APFS_NXI(sb)->nx_raw;
-	struct apfs_node *vnode;
-	struct apfs_omap_phys *msb_omap_raw;
-	struct buffer_head *bh;
-	struct apfs_omap *omap = NULL;
-	u64 msb_omap, vb;
-	int i;
-	int err = 0;
-
-	/* Get the container's object map */
-	msb_omap = le64_to_cpu(msb_raw->nx_omap_oid);
-	bh = apfs_sb_bread(sb, msb_omap);
-	if (!bh) {
-		apfs_err(sb, "unable to read container object map");
-		return -EIO;
-	}
-	msb_omap_raw = (struct apfs_omap_phys *)bh->b_data;
-
-	/* Get the Volume Block */
-	vb = le64_to_cpu(msb_omap_raw->om_tree_oid);
-	msb_omap_raw = NULL;
-	brelse(bh);
-	bh = NULL;
-	vnode = apfs_read_node(sb, vb, APFS_OBJ_PHYSICAL, false /* write */);
-	if (IS_ERR(vnode)) {
-		apfs_err(sb, "unable to read volume block");
-		return PTR_ERR(vnode);
-	}
-
-	omap = apfs_alloc_omap();
-	if (!omap) {
-		err = -ENOMEM;
-		goto fail;
-	}
-	omap->omap_root = vnode;
-
-	/* Iterate through the checkpoint superblocks and add the used blocks */
-	*count = 0;
-	for (i = 0; i < APFS_NX_MAX_FILE_SYSTEMS; i++) {
-		struct apfs_superblock *vsb_raw;
-		u64 vol_id;
-		u64 vol_bno;
-
-		vol_id = le64_to_cpu(msb_raw->nx_fs_oid[i]);
-		if (vol_id == 0) /* All volumes have been checked */
-			break;
-		err = apfs_omap_lookup_newest_block(sb, omap, vol_id, &vol_bno, false /* write */);
-		if (err) {
-			apfs_err(sb, "omap lookup failed for vol id 0x%llx", vol_id);
-			break;
-		}
-
-		bh = apfs_sb_bread(sb, vol_bno);
-		if (!bh) {
-			err = -EIO;
-			apfs_err(sb, "unable to read volume superblock");
-			break;
-		}
-		vsb_raw = (struct apfs_superblock *)bh->b_data;
-		*count += le64_to_cpu(vsb_raw->apfs_fs_alloc_count);
-		brelse(bh);
-	}
-
-fail:
-	kfree(omap);
-	apfs_node_free(vnode);
-	return err;
-}
-
 static int apfs_statfs(struct dentry *dentry, struct kstatfs *buf)
 {
 	struct super_block *sb = dentry->d_sb;
@@ -1024,7 +1068,7 @@ static int apfs_statfs(struct dentry *dentry, struct kstatfs *buf)
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
 	struct apfs_nx_superblock *msb_raw;
 	struct apfs_superblock *vol;
-	u64 fsid, used_blocks = 0;
+	u64 fsid, free_blocks;
 	int err;
 
 	down_read(&nxi->nx_big_sem);
@@ -1037,11 +1081,16 @@ static int apfs_statfs(struct dentry *dentry, struct kstatfs *buf)
 
 	/* Volumes share the whole disk space */
 	buf->f_blocks = le64_to_cpu(msb_raw->nx_block_count);
-	err = apfs_count_used_blocks(sb, &used_blocks);
+	/*
+	 * It takes some work to retrieve the free block count because we
+	 * can't assume that the spaceman has been read yet. It would be
+	 * cleaner if we always did that on first mount (TODO).
+	 */
+	err = apfs_spaceman_get_free_blkcnt(sb, &free_blocks);
 	if (err)
 		goto fail;
-	buf->f_bfree = buf->f_blocks - used_blocks;
-	buf->f_bavail = buf->f_bfree; /* I don't know any better */
+	buf->f_bfree = free_blocks;
+	buf->f_bavail = free_blocks;
 
 	/* The file count is only for the mounted volume */
 	buf->f_files = le64_to_cpu(vol->apfs_num_files) +
@@ -1084,20 +1133,21 @@ static int apfs_show_options(struct seq_file *seq, struct dentry *root)
 						     sbi->s_gid));
 	if (nxi->nx_flags & APFS_CHECK_NODES)
 		seq_puts(seq, ",cknodes");
+	if (nxi->nx_tier2_info)
+		seq_printf(seq, ",tier2=%s", nxi->nx_tier2_info->blki_path);
 
 	return 0;
 }
 
 int apfs_sync_fs(struct super_block *sb, int wait)
 {
-	struct apfs_max_ops maxops = {0};
 	int err;
 
 	/* TODO: actually start the commit and return without waiting? */
 	if (wait == 0)
 		return 0;
 
-	err = apfs_transaction_start(sb, maxops);
+	err = apfs_transaction_start(sb, APFS_TRANS_SYNC);
 	if (err)
 		return err;
 	APFS_SB(sb)->s_nxi->nx_transaction.t_state |= APFS_NX_TRANS_FORCE_COMMIT;
@@ -1142,7 +1192,7 @@ static const struct super_operations apfs_sops = {
 };
 
 enum {
-	Opt_readwrite, Opt_cknodes, Opt_uid, Opt_gid, Opt_vol, Opt_snap, Opt_err,
+	Opt_readwrite, Opt_cknodes, Opt_uid, Opt_gid, Opt_vol, Opt_snap, Opt_tier2, Opt_err,
 };
 
 static const match_table_t tokens = {
@@ -1152,6 +1202,7 @@ static const match_table_t tokens = {
 	{Opt_gid, "gid=%u"},
 	{Opt_vol, "vol=%u"},
 	{Opt_snap, "snap=%s"},
+	{Opt_tier2, "tier2=%s"},
 	{Opt_err, NULL}
 };
 
@@ -1193,6 +1244,11 @@ static int parse_options(struct super_block *sb, char *options)
 	/* Set default values before parsing */
 	nx_flags = 0;
 
+#ifdef CONFIG_APFS_RW_ALWAYS
+	/* Still risky, but some packagers want writable mounts by default */
+	nx_flags |= APFS_READWRITE;
+#endif
+
 	if (!options)
 		goto out;
 
@@ -1239,9 +1295,15 @@ static int parse_options(struct super_block *sb, char *options)
 			break;
 		case Opt_vol:
 		case Opt_snap:
+		case Opt_tier2:
 			/* Already read early on mount */
 			break;
 		default:
+			/*
+			 * We should have already checked the mount options in
+			 * apfs_preparse_options(), so this is a bug.
+			 */
+			apfs_alert(sb, "invalid mount option %s", p);
 			return -EINVAL;
 		}
 	}
@@ -1271,6 +1333,7 @@ static int apfs_check_nx_features(struct super_block *sb)
 {
 	struct apfs_nx_superblock *msb_raw = NULL;
 	u64 features;
+	bool fusion;
 
 	msb_raw = APFS_NXI(sb)->nx_raw;
 	if (!msb_raw) {
@@ -1283,10 +1346,22 @@ static int apfs_check_nx_features(struct super_block *sb)
 		apfs_warn(sb, "unknown incompatible container features (0x%llx)", features);
 		return -EINVAL;
 	}
-	if (features & APFS_NX_INCOMPAT_FUSION) {
-		apfs_warn(sb, "fusion drives are not supported");
+
+	fusion = features & APFS_NX_INCOMPAT_FUSION;
+	if (fusion && !APFS_NXI(sb)->nx_tier2_info) {
+		apfs_warn(sb, "fusion drive - please use the \"tier2\" mount option");
 		return -EINVAL;
 	}
+	if (!fusion && APFS_NXI(sb)->nx_tier2_info) {
+		apfs_warn(sb, "not a fusion drive - what's the second disk for?");
+		return -EINVAL;
+	}
+	if (fusion) {
+		if (!sb_rdonly(sb)) {
+			apfs_warn(sb, "writes to fusion drives not yet supported");
+			sb->s_flags |= SB_RDONLY;
+		}
+	}
 
 	features = le64_to_cpu(msb_raw->nx_readonly_compatible_features);
 	if (features & ~APFS_NX_SUPPORTED_ROCOMPAT_MASK) {
@@ -1323,8 +1398,15 @@ static int apfs_check_vol_features(struct super_block *sb)
 		return -EINVAL;
 	}
 	if (features & APFS_INCOMPAT_DATALESS_SNAPS) {
-		apfs_warn(sb, "snapshots with no data are not supported");
-		return -EINVAL;
+		/*
+		 * I haven't encountered dataless snapshots myself yet (TODO).
+		 * I'm not even sure what they are, so be safe.
+		 */
+		if (!sb_rdonly(sb)) {
+			apfs_warn(sb, "writes to volumes with dataless snapshots not yet supported");
+			return -EINVAL;
+		}
+		apfs_warn(sb, "volume has dataless snapshots");
 	}
 	if (features & APFS_INCOMPAT_ENC_ROLLED) {
 		apfs_warn(sb, "encrypted volumes are not supported");
@@ -1387,13 +1469,15 @@ static int apfs_check_vol_features(struct super_block *sb)
 static int apfs_setup_bdi(struct super_block *sb)
 {
 	struct apfs_nxsb_info *nxi = APFS_NXI(sb);
+	struct apfs_blkdev_info *bd_info = NULL;
 	struct backing_dev_info *bdi_dev = NULL, *bdi_sb = NULL;
 	int err;
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0) || (defined(RHEL_RELEASE) && LINUX_VERSION_CODE == KERNEL_VERSION(5, 14, 0))
-	bdi_dev = nxi->nx_bdev->bd_disk->bdi;
+	bd_info = nxi->nx_blkdev_info;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0) || RHEL_VERSION_GE(9, 0)
+	bdi_dev = bd_info->blki_bdev->bd_disk->bdi;
 #else
-	bdi_dev = nxi->nx_bdev->bd_bdi;
+	bdi_dev = bd_info->blki_bdev->bd_bdi;
 #endif
 
 	err = super_setup_bdi(sb);
@@ -1611,8 +1695,10 @@ static int apfs_set_super(struct super_block *sb, void *data)
 	 * This is the actual device number, shared by all volumes and
 	 * snapshots. It gets reported by the mountinfo file, and it seems that
 	 * udisks uses it to decide if a device is mounted, so it must be set.
+	 *
+	 * TODO: does this work for fusion drives?
 	 */
-	sb->s_dev = nxi->nx_bdev->bd_dev;
+	sb->s_dev = nxi->nx_blkdev_info->blki_bdev->bd_dev;
 
 	sb->s_fs_info = sbi;
 	return 0;
@@ -1638,6 +1724,79 @@ static int apfs_lookup_bdev(const char *pathname, dev_t *dev)
 #endif
 }
 
+/**
+ * apfs_blkdev_setup - Open a block device and set its info struct
+ * @info_p:	info struct to set
+ * @dev_name:	path name for the block device to open
+ * @mode:	FMODE_* mask
+ *
+ * Returns 0 on success, or a negative error code in case of failure.
+ */
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0)
+static int apfs_blkdev_setup(struct apfs_blkdev_info **info_p, const char *dev_name, blk_mode_t mode)
+#else
+static int apfs_blkdev_setup(struct apfs_blkdev_info **info_p, const char *dev_name, fmode_t mode)
+#endif
+{
+	struct apfs_blkdev_info *info = NULL;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 9, 0) || RHEL_VERSION_GE(9, 5)
+	struct file *file = NULL;
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0)
+	struct bdev_handle *handle = NULL;
+#endif
+	struct block_device *bdev = NULL;
+	int ret;
+
+	info = kzalloc(sizeof(*info), GFP_KERNEL);
+	if (!info)
+		return -ENOMEM;
+	info->blki_path = kstrdup(dev_name, GFP_KERNEL);
+	if (!info->blki_path) {
+		ret = -ENOMEM;
+		goto fail;
+	}
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 9, 0) || RHEL_VERSION_GE(9, 5)
+	file = bdev_file_open_by_path(dev_name, mode, &apfs_fs_type, NULL);
+	if (IS_ERR(file)) {
+		ret = PTR_ERR(file);
+		goto fail;
+	}
+	info->blki_bdev_file = file;
+	bdev = file_bdev(file);
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0)
+	handle = bdev_open_by_path(dev_name, mode, &apfs_fs_type, NULL);
+	if (IS_ERR(handle)) {
+		ret = PTR_ERR(handle);
+		goto fail;
+	}
+	info->blki_bdev_handle = handle;
+	bdev = handle->bdev;
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0) || RHEL_VERSION_GE(9, 4)
+	bdev = blkdev_get_by_path(dev_name, mode, &apfs_fs_type, NULL);
+#else
+	bdev = blkdev_get_by_path(dev_name, mode, &apfs_fs_type);
+#endif
+
+	if (IS_ERR(bdev)) {
+		ret = PTR_ERR(bdev);
+		goto fail;
+	}
+	info->blki_bdev = bdev;
+#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 5, 0) && !RHEL_VERSION_GE(9, 4)
+	info->blki_mode = mode;
+#endif
+	*info_p = info;
+	return 0;
+
+fail:
+	kfree(info->blki_path);
+	info->blki_path = NULL;
+	kfree(info);
+	info = NULL;
+	return ret;
+}
+
 /**
  * apfs_attach_nxi - Attach container sb info to a volume's sb info
  * @sbi:	new superblock info structure for the volume to be mounted
@@ -1652,7 +1811,7 @@ static int apfs_attach_nxi(struct apfs_sb_info *sbi, const char *dev_name, blk_m
 static int apfs_attach_nxi(struct apfs_sb_info *sbi, const char *dev_name, fmode_t mode)
 #endif
 {
-	struct apfs_nxsb_info *nxi;
+	struct apfs_nxsb_info *nxi = NULL;
 	dev_t dev = 0;
 	int ret;
 
@@ -1664,49 +1823,28 @@ static int apfs_attach_nxi(struct apfs_sb_info *sbi, const char *dev_name, fmode
 
 	nxi = apfs_nx_find_by_dev(dev);
 	if (!nxi) {
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 9, 0)
-		struct file *file = NULL;
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0)
-		struct bdev_handle *handle = NULL;
-#endif
-		struct block_device *bdev = NULL;
-
 		nxi = kzalloc(sizeof(*nxi), GFP_KERNEL);
 		if (!nxi) {
 			ret = -ENOMEM;
 			goto out;
 		}
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 9, 0)
-		file = bdev_file_open_by_path(dev_name, mode, &apfs_fs_type, NULL);
-		if (IS_ERR(file)) {
-			kfree(nxi);
-			ret = PTR_ERR(file);
-			goto out;
-		}
-		nxi->nx_bdev_file = file;
-		bdev = file_bdev(file);
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 8, 0)
-		handle = bdev_open_by_path(dev_name, mode, &apfs_fs_type, NULL);
-		if (IS_ERR(handle)) {
-			kfree(nxi);
-			ret = PTR_ERR(handle);
-			goto out;
-		}
-		nxi->nx_bdev_handle = handle;
-		bdev = handle->bdev;
-#elif LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0)
-		bdev = blkdev_get_by_path(dev_name, mode, &apfs_fs_type, NULL);
-#else
-		bdev = blkdev_get_by_path(dev_name, mode, &apfs_fs_type);
-#endif
-		if (IS_ERR(bdev)) {
-			kfree(nxi);
-			ret = PTR_ERR(bdev);
+		ret = apfs_blkdev_setup(&nxi->nx_blkdev_info, dev_name, mode);
+		if (ret)
 			goto out;
+
+		if (sbi->s_tier2_path) {
+			ret = apfs_blkdev_setup(&nxi->nx_tier2_info, sbi->s_tier2_path, mode);
+			if (ret) {
+				apfs_blkdev_cleanup(nxi->nx_blkdev_info);
+				nxi->nx_blkdev_info = NULL;
+				goto out;
+			}
+			/* We won't need this anymore, so why waste memory? */
+			kfree(sbi->s_tier2_path);
+			sbi->s_tier2_path = NULL;
 		}
 
-		nxi->nx_bdev = bdev;
 		init_rwsem(&nxi->nx_big_sem);
 		list_add(&nxi->nx_list, &nxs);
 		INIT_LIST_HEAD(&nxi->vol_list);
@@ -1718,6 +1856,10 @@ static int apfs_attach_nxi(struct apfs_sb_info *sbi, const char *dev_name, fmode
 	++nxi->nx_refcnt;
 	ret = 0;
 out:
+	if (ret) {
+		kfree(nxi);
+		nxi = NULL;
+	}
 	mutex_unlock(&nxs_mutex);
 	return ret;
 }
@@ -1740,6 +1882,7 @@ static int apfs_preparse_options(struct apfs_sb_info *sbi, char *options)
 	/* Set default values before parsing */
 	sbi->s_vol_nr = 0;
 	sbi->s_snap_name = NULL;
+	sbi->s_tier2_path = NULL;
 
 	if (!options)
 		return 0;
@@ -1770,6 +1913,14 @@ static int apfs_preparse_options(struct apfs_sb_info *sbi, char *options)
 				goto out;
 			}
 			break;
+		case Opt_tier2:
+			kfree(sbi->s_tier2_path);
+			sbi->s_tier2_path = match_strdup(&args[0]);
+			if (!sbi->s_tier2_path) {
+				err = -ENOMEM;
+				goto out;
+			}
+			break;
 		case Opt_readwrite:
 		case Opt_cknodes:
 		case Opt_uid:
@@ -1777,6 +1928,7 @@ static int apfs_preparse_options(struct apfs_sb_info *sbi, char *options)
 			/* Not needed for sget(), will be read later */
 			break;
 		default:
+			apfs_warn(NULL, "invalid mount option %s", p);
 			err = -EINVAL;
 			goto out;
 		}
@@ -1795,7 +1947,8 @@ static struct dentry *apfs_mount(struct file_system_type *fs_type, int flags,
 {
 	struct super_block *sb;
 	struct apfs_sb_info *sbi;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0)
+	struct apfs_blkdev_info *bd_info = NULL, *tier2_info = NULL;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0) || RHEL_VERSION_GE(9, 4)
 	blk_mode_t mode = sb_open_mode(flags);
 #else
 	fmode_t mode = FMODE_READ | FMODE_EXCL;
@@ -1814,7 +1967,7 @@ static struct dentry *apfs_mount(struct file_system_type *fs_type, int flags,
 	if (sbi->s_snap_name)
 		flags |= SB_RDONLY;
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 5, 0)
+#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 5, 0) && !RHEL_VERSION_GE(9, 4)
 	if (!(flags & SB_RDONLY))
 		mode |= FMODE_WRITE;
 #endif
@@ -1830,11 +1983,14 @@ static struct dentry *apfs_mount(struct file_system_type *fs_type, int flags,
 		goto out_unmap_super;
 	}
 
+	bd_info = APFS_NXI(sb)->nx_blkdev_info;
+	tier2_info = APFS_NXI(sb)->nx_tier2_info;
+
 	/*
 	 * I'm doing something hacky with s_dev inside ->kill_sb(), so I want
 	 * to find out as soon as possible if I messed it up.
 	 */
-	WARN_ON(sb->s_dev != APFS_NXI(sb)->nx_bdev->bd_dev);
+	WARN_ON(sb->s_dev != bd_info->blki_bdev->bd_dev);
 
 	if (sb->s_root) {
 		if ((flags ^ sb->s_flags) & SB_RDONLY) {
@@ -1846,18 +2002,27 @@ static struct dentry *apfs_mount(struct file_system_type *fs_type, int flags,
 		apfs_free_main_super(sbi);
 		kfree(sbi->s_snap_name);
 		sbi->s_snap_name = NULL;
+		kfree(sbi->s_tier2_path);
+		sbi->s_tier2_path = NULL;
 		kfree(sbi);
 		sbi = NULL;
 	} else {
+		if (!sbi->s_snap_name && !tier2_info)
+			snprintf(sb->s_id, sizeof(sb->s_id), "%pg:%u", bd_info->blki_bdev, sbi->s_vol_nr);
+		else if (!tier2_info)
+			snprintf(sb->s_id, sizeof(sb->s_id), "%pg:%u:%s", bd_info->blki_bdev, sbi->s_vol_nr, sbi->s_snap_name);
+		else if (!sbi->s_snap_name)
+			snprintf(sb->s_id, sizeof(sb->s_id), "%pg+%pg:%u", bd_info->blki_bdev, tier2_info->blki_bdev, sbi->s_vol_nr);
+		else
+			snprintf(sb->s_id, sizeof(sb->s_id), "%pg+%pg:%u:%s", bd_info->blki_bdev, tier2_info->blki_bdev, sbi->s_vol_nr, sbi->s_snap_name);
 		error = apfs_read_main_super(sb);
 		if (error) {
 			deactivate_locked_super(sb);
 			return ERR_PTR(error);
 		}
-#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 5, 0)
+#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 5, 0) && !RHEL_VERSION_GE(9, 4)
 		sb->s_mode = mode;
 #endif
-		snprintf(sb->s_id, sizeof(sb->s_id), "%xg", sb->s_dev);
 		error = apfs_fill_super(sb, data, flags & SB_SILENT ? 1 : 0);
 		if (error) {
 			deactivate_locked_super(sb);
@@ -1872,6 +2037,7 @@ out_unmap_super:
 	apfs_free_main_super(sbi);
 out_free_sbi:
 	kfree(sbi->s_snap_name);
+	kfree(sbi->s_tier2_path);
 	kfree(sbi);
 	return ERR_PTR(error);
 }
@@ -1894,6 +2060,8 @@ static void apfs_free_sb_info(struct super_block *sb)
 
 	kfree(sbi->s_snap_name);
 	sbi->s_snap_name = NULL;
+	kfree(sbi->s_tier2_path);
+	sbi->s_tier2_path = NULL;
 	if (sbi->s_dflt_pfk)
 		kfree(sbi->s_dflt_pfk);
 	kfree(sbi);
